package lockmgr;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class LockManager implements ILockManager{
	
	private Hashtable<String, UpgradableLock> lockTable;
	private Hashtable<Integer, Set<UpgradableLock>> lockedObjects;
	private Hashtable<String, Set<Integer>> lockHolders;
	
	private WaitingGraph wg;
	
	private final Lock lockDataStructure1 = new ReentrantLock(true);
	
	
	
	public LockManager() {
		this.lockTable = new Hashtable<String, UpgradableLock>();
		this.lockedObjects = new Hashtable<Integer, Set<UpgradableLock>>();
		this.lockHolders = new Hashtable<String, Set<Integer>>();
		
		this.wg = new WaitingGraph();
	}

	public boolean lock(int xid, String strData, LockType lockType)
	throws DeadlockException {
		UpgradableLock lock = null;
		Set<Integer> lh = null;
		
		//****** ADJUST DATA STRUCTURES ********
		synchronized(lockDataStructure1){
			//check for conflict
			lock = lockTable.get(strData);
			if(lock == null){
				//no conflict since no lock exists at all
				//add lock entry
				lock = new UpgradableLock();								
				lockTable.put(strData, lock);																												
			}
			
			lh = lockHolders.get(strData);
			if(lh == null){
				lh = Collections.synchronizedSet(new HashSet<Integer>());
				lockHolders.put(strData, lh);
			}
		}		
		
		//remember locks held by transaction
		//hashtable is already synchronized and we assume that each thread is
		//associated with a unique xid --> no explicit synchronization necessary
		Set<UpgradableLock> lo = lockedObjects.get(xid);
		if(lo == null){
			lo = Collections.synchronizedSet(new HashSet<UpgradableLock>());	
			lockedObjects.put(xid, lo);
		}
		
		//****** TRY TO ACQUIRE LOCKS ********
		switch(lockType){
		
		//READ
		case READ:
			
			if(lock.holdsReadlock())
				return true;
			
			//remember that this transaction wants to acquire a lock
			lh.add(xid);
			if(!lock.tryReadLock()){
				//lock could not be acquired --> deadlock detection
				if(deadlock(xid, lh))
					throw new DeadlockException(xid, "rl(" + strData + ")");
				
				lock.readLock();
			}			
			//read lock has been acquired				
			lo.add(lock);				

			break;
			
		//WRITE
		case WRITE:		
			
			if(lock.holdsWritelock())
				return true;
			
			//lock upgrade?
			if(!lock.holdsReadlock()){
				//NO UPGRADE				
				if(!lock.tryWriteLock()){
					if(deadlock(xid, lh)){						
						throw new DeadlockException(xid, "wl(" + strData + ")");
					}
					lock.writeLock();
				}											
				lh.add(xid);
				lo.add(lock);							
			}
			else{													
				//LOCK UPGRADE														
				if(!lock.tryWriteLock()){
					if(deadlock(xid, lh)){
						throw new DeadlockException(xid, "wl(" + strData + ")");
					}
					lock.upgradeLock();						
				}							
				lh.add(xid);
				lo.add(lock);
			}
					
		}				

		return true;
	}
	
	private boolean deadlock(int xid, Set<Integer> lockHolders){
		//add new edge from this transaction to the transaction
		//holding a read/write lock on the object
		
		//a transaction that upgrades its lock from read to write
		//is NOT waiting for itself! --> exclude it from lockHoldes
		Set<Integer> lockHoldersCopy = new HashSet<Integer>();
		for (Integer integer : lockHolders) {
			if(integer != xid)
				lockHoldersCopy.add(integer);
		}
		synchronized(wg){
			if(!wg.addEdges(xid, lockHoldersCopy)){
				return true;
			}
		}
		
		
		return false;
	}
	
	public boolean unlockAll(int xid) {
		
		synchronized(lockDataStructure1){
			Set<UpgradableLock> locks = lockedObjects.get(xid);
			if(locks != null){
				for (UpgradableLock lock : locks) {				
					try {
						lock.unlock();					
					} catch (IllegalMonitorStateException e) {
						throw new RuntimeException(e);
					}
				}
				lockedObjects.remove(xid);
				for (Set<Integer> lh : lockHolders.values()) {
					lh.remove(xid);
				}
				wg.removeTransaction(xid);
			}
		}
		
		return true;
	}
	
	/**
	 * A simple graph data structure that allows to check for cycles.
	 * @author bross
	 *
	 */
	private class WaitingGraph{
		
		private final Hashtable<Integer, Set<Integer>> adjacencyList =
			new Hashtable<Integer, Set<Integer>>();				

		public synchronized boolean addEdges(int source, Set<Integer> targets){
			
			Set<Integer> set = adjacencyList.get(source);
			if(set == null){
				adjacencyList.put(source,targets);
			}
			else
				set.addAll(targets);
			
			return !containsCycle();
		}
		
		public synchronized void removeTransaction(int xid){
			adjacencyList.remove(xid);
			for (Set<Integer> adjacentVertices : adjacencyList.values()) {
				adjacentVertices.remove(xid);
			}
		}
		
		private synchronized boolean containsCycle(){
			
			//depth first search			
			HashMap<Integer, Integer> status = new HashMap<Integer, Integer>();
						
			for (Integer integer : adjacencyList.keySet()) {
				if(!status.containsKey(integer)){
					try {
						DFSVisit(integer, status);
					} catch (Exception e) {
						return true;
					}					
				}
											
			}
			return false;
		}
		
		private void DFSVisit(int v, HashMap<Integer, Integer> status) throws Exception{
			status.put(v, 1);
			if(adjacencyList.get(v) != null){										
				for (Integer adjacent : adjacencyList.get(v)) {
					if(!status.containsKey(adjacent))
						DFSVisit(adjacent, status);
					else if(status.get(adjacent) == 1) //detect cycle
						throw new Exception();
				}
			}
			
			status.put(v, 2);
		}
	}
}
