package transaction;

import java.rmi.Remote;
import java.rmi.RemoteException;



/**
 * /** 
 * Interface for a simple Resource Manager for the Distributed Travel
 * Reservation System. A single resource manager can be used to manage
 * arbitrary resources, e.g. hotels, rental cars or flights.
 * <p>
 * Failure reporting is done using two pieces, exceptions and
 * return values.  Exceptions are used for systemy things - like
 * transactions that were forced to abort, or don't exist.  Return
 * values are used for operations that would affect the consistency of
 * the database, like the reservation of a data item that the database
 * does not know about.
 * <p>
 * If there is a return value and you're not sure how it would
 * be used in your implementation, ignore it.  Return
 * values in the interface are used generously to allow flexibility in
 * implementation.
 *
 * @author bross
 *
 */
public interface IResourceManager extends Remote{
	
	public static final String RMI_NAME_HOTEL_RM = "hotelRM";
	public static final String RMI_NAME_FLIGHT_RM = "flightRM";
	public static final String RMI_NAME_CAR_RM = "carRM";
	
	public void start(String workingdir, String rmiName) throws RemoteException;
	
	public boolean connectTM(String rmiName) throws RemoteException;

    //////////
    // TRANSACTION INTERFACE
    //////////
    /**
     * @deprecated Only distributed transaction should by allowed from now on.
     * Begin a new local transaction, and return its transaction id.
     *
     * @return A unique local transaction ID > 0.    
     * 
     * @throws ShutdownException thrown if no new transactions can be accepted
     * due to shutdown of the resource manager
     */
    public int begin() throws ShutdownException, RemoteException;
    
    /**
     * Begin a new distributed transaction. Part of starting a new
     * distributed transaction is to call the transaction manager's
     * enlist() method in order to register this RM as a participant
     * of the given distributed transaction.
     *
     * @param globalXID the id of the distributed transaction which
     * will be the parent of this local subtransaction. Use this id to
     * enlist this RM at the transaction manager.
     * @return true if successful, false otherwise 
     * 
     * @throws ShutdownException thrown if no new transactions can be accepted
     * due to shutdown of the resource manager
     * @throws InvalidTransactionException if a distributed transaction with
     * this globalXID has already been started at this RM
     */
    public boolean beginDistributed(int globalXID)
    throws ShutdownException, RemoteException, InvalidTransactionException;
    
    /**
     * This method is part of the 2PC protocol for atomic committment of
     * distributed transactions.
     * @param globalXID the transaction ID of the distributed transaction
     * which needs to be prepared for commitment
     * @return true if prepare was successful, false otherwise
     * @throws InvalidTransactionException
     * @throws RemoteException
     */
    public boolean prepare(int globalXID) throws InvalidTransactionException, RemoteException;
    
    /**
     * Commit transaction.
     *
     * @param xid id of transaction to be committed.
     * @return true on success, false on failure.   
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public boolean commit(int xid) 
	throws TransactionAbortedException,
	       InvalidTransactionException, RemoteException;
    /**
     * Abort transaction.
     *
     * @param xid id of transaction to be aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public void abort(int xid) 
	throws InvalidTransactionException, RemoteException;
    
    
    //////////
    // ADMINISTRATIVE INTERFACE
    //////////
    
    /**
     * Add a new customer to the database. A customer is
     * identified by his unique customer name (e.g. Smith)
     * @param xid the id of the transaction requesting this operation
     * @param custName the unique customer name
     * @return true if the operation succeeded, false otherwise
     * (e.g. customer already exists or name is null, etc.)
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public boolean newCustomer(int xid, String custName) 
	throws TransactionAbortedException,
	       InvalidTransactionException, RemoteException;
    
    /**
     * Query the bill of a customer.
     * @param xid the id of the transaction requesting this operation
     * @param custName the unique customer name whose bill should be calculated
     * @return the sum of costs accumulated due to reservations of this customer,
     * -1 if the customer does not exist
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public int queryCustomerBill(int xid, String custName)
    throws TransactionAbortedException,
	   InvalidTransactionException, RemoteException;
    
    /**
     * Add a new resource (e.g. hotel) to the database. A resource
     * is identified by a unique name (e.g. HotelA).
     * @param xid the id of the transaction requesting this operation
     * @param rid the unique identifier of the resource which is to be added (e.g. HotelA)
     * @param num the total capacity of this resource 
     * @param price the price a customer needs to pay for reserving this resource
     * @return true if the operation succeeded, false otherwise (e.g. resource already exists)
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public boolean addResource(int xid, String rid, int num, int price)
    throws TransactionAbortedException, InvalidTransactionException, RemoteException;
    
    /**
     * Update the attributes (total capacity and price) of a resource.
     * @param xid the id of the transaction requesting this operation
     * @param rid the unique identifier of the resource which is to be added (e.g. HotelA)
     * @param num a number indicating the change in total capacity, negative numbers decrease
     * capacity, positive values increase capacity  
     * @param price the price a customer needs to pay for reserving this resource
     * @return true if the operation succeeded, false otherwise
     * (e.g. decrease > total capacity - reservations)
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public boolean updateResource(int xid, String rid, int num, int price)
    throws TransactionAbortedException, InvalidTransactionException, RemoteException;
    
    /**
     * Query the availability of a resource.
     * @param xid the id of the transaction requesting this operation
     * @param rid the unique identifier of the resource
     * @return the current availability (i.e. total capacity - reservations),
     * -1 if the resource does not exist
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public int queryResourceAvailability(int xid, String rid)
    throws TransactionAbortedException, InvalidTransactionException, RemoteException;
    
    /**
     * 
     * Query the price of a resource.
     * @param xid the id of the transaction requesting this operation
     * @param rid the unique identifier of the resource
     * @return the price of the resource, -1 if the resource does not exist
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public int queryResourcePrice(int xid, String rid)
    throws TransactionAbortedException, InvalidTransactionException, RemoteException;
    
    /**
     * Reserve a resource for a given customer. The availability
     * of the resource is decreased by one.
    * @param xid the id of the transaction requesting this operation
     * @param rid the unique identifier of the resource
     * @param custName the unique name of the customer
     * @return true if the operation succeeded, false otherwise
     * (e.g. customer or resource do not exist, or no availability)
     * @throws TransactionAbortedException if transaction was aborted.
     * @throws InvalidTransactionException if transaction id is invalid.
     */
    public boolean reserveResource(int xid, String rid, String custName)
    throws TransactionAbortedException, InvalidTransactionException, RemoteException;
    

    //////////
    // TECHNICAL/TESTING INTERFACE
    //////////
    /** 
     * Shutdown gracefully. Stop accepting new transactions, wait for
     * running transactions to terminate, and clean up disk state.
     * When this RM restarts, it should not attempt to recover its
     * state if the client called shutdown to terminate it.
     *
     * @return true on success, false on failure.
     */
    public boolean shutdown() throws RemoteException;
    
    /**
     * immediately closes the ResourceManager by calling System.exit(-1)
     */
    public void dieNow() throws RemoteException;
    
    /**
     * Closes ResourceManager by calling System.exit(-1) right before
     * switching the pointer to the shadow copy file.
     */
    public void dieBeforePointerSwitch() throws RemoteException;
    
    /**
     * Closes ResourceManager by calling System.exit(-1) right after
     * switching the pointer to the shadow copy file.
     */
    public void dieAfterPointerSwitch() throws RemoteException;
}
