package transaction;


/**
 * This interface describes the booking system of a travel
 * agency. A travel agency uses several autonomous resource
 * managers to book an itinerary. This interface defines an agency
 * that can book hotels, flights and rental cars. An implementation
 * must use a separate resource manager for each booking type. That
 * is, one RM is responsible for hotels, one for flights and another
 * one for rental cars.
 * 
 * Transactions are started implicitly by calling one of the RMs. At most
 * one transaction is active in one TravelAgency terminal. However,
 * there may exist more than one TravelAgency terminal in parallel.
 * 
 * Each transaction started implicitly by a travel agency terminal
 * is always a <b>distributed transaction</b>. Use the transaction manager's
 * begin() method to retrieve a unique global ID for the distributed
 * transaction.
 * 
 * Use the RM.beginDistributed(int globalXID) method to start
 * an new subtransaction.
 *  
 * @author bross
 *
 */
public interface ITravelAgency {
	
	// ***************** RESERVATION INTERFACE ******************
	
	
	/**
	 * Reserve a resource. Uses the RM responsible for the given
	 * type of resource to book.
	 * If the customer is not known to the RM, add a new customer.
	 * Only book the resource if it is available.
	 * @param resourceType the type of resource (e.g. HOTEL)
	 * @param rid the unique name of the hotel
	 * @param customerName the unique customer name
	 * @return true if successfull, false otherwise
	 */
	public boolean reserve(ResourceType resourceType, String rid, String customerName);
	
	/**
	 * Convenience method for reserving
	 * a whole itinerary. Uses all three RMs to book.
	 * If the customer is not known to one of the RM, add a new customer to this RM.
	 * Only book the resources if they are available.
	 * @param hotelName the unique hotel name
	 * @param flightNumber the unique flight number
	 * @param agencyName the unique name of the car rental agency
	 * @param customerName the unique customer name
	 * @return true if successfull, false otherwise
	 */
	public boolean reserveItinerary(String hotelName, String flightNumber, String agencyName, String customerName);
	
	
	// ***************** QUERY INTERFACE ******************
	
	/**
	 * Query the price of a resource.
	 * @param resourceType the type of resource (e.g. HOTEL)
	 * @param rid the unique name of the resource
	 * @return the price of the resource or a value < 0 if resource does not exist
	 */
	public int queryPrice(ResourceType resourceType, String rid);
	
	/**
	 * Query the availability of a resource.
	 * @param resourceType the type of resource (e.g. HOTEL)
	 * @param rid the unique name of the resource
	 * @return remaining capacity of the resource or value < 0 if resource does not exist
	 */
	public int queryAvailability(ResourceType resourceType, String rid);
	
	/**
	 * Calculate the bill for a certain customer. Queries all three RMs to
	 * sum up the bill.
	 * @param customerName the unique customer name
	 * @return the sum over all costs for hotel, flight and car reservations
	 */
	public int queryBill(String customerName);
	
	
	// ***************** ADMINISTRATION INTERFACE ******************
	
	/**
     * Add a new customer to the database. A customer is
     * identified by his unique customer name (e.g. Smith)
     * @param resourceType the type of resource (e.g. HOTEL)
     * @param custName the unique customer name
     * @return true if the operation succeeded, false otherwise
     * (e.g. customer already exists or name is null, etc.)
     */
    public boolean newCustomer(ResourceType resourceType, String custName);
    
    /**
     * Add a new resource (e.g. hotel) to the database. A resource
     * is identified by a unique name (e.g. HotelA).
     * @param resourceType the type of resource (e.g. HOTEL)
     * @param rid the unique identifier of the resource which is to be added (e.g. HotelA)
     * @param num the total capacity of this resource 
     * @param price the price a customer needs to pay for reserving this resource
     * @return true if the operation succeeded, false otherwise (e.g. resource already exists)
     */
    public boolean addResource(ResourceType resourceType,  String rid, int num, int price);
    
    /**
     * Update the attributes (total capacity and price) of a resource.
     * @param resourceType the type of resource (e.g. HOTEL)
     * @param rid the unique identifier of the resource which is to be added (e.g. HotelA)
     * @param num a number indicating the change in total capacity, negative numbers decrease
     * capacity, positive values increase capacity  
     * @param price the price a customer needs to pay for reserving this resource
     * @return true if the operation succeeded, false otherwise
     * (e.g. decrease > total capacity - reservations)
     */
    public boolean updateResource(ResourceType resourceType, String rid, int num, int price);
	
	
	// ***************** TRANSACTION CONTROL INTERFACE ******************
	
	/**
	 * Commit the current transaction. Uses the Transaction Manager to commit
	 * the distributed transaction.
	 * @return true
	 * @throws InvalidTransactionException
	 * @throws TransactionAbortedException
	 */
	public boolean commit() throws TransactionAbortedException, InvalidTransactionException;
	
	/**
	 * 
	 * Abort the current transaction. Uses the Transaction Manager to abort
	 * the distributed transaction.
	 * @return true
	 * @throws TransactionAbortedException
	 * @throws InvalidTransactionException
	 */
	public boolean abort() throws TransactionAbortedException, InvalidTransactionException;

}
